'''
Use python version 3.4
'''

__all__ = ['python']

class Python:

    name = ''
    pip = None

    def __init__(self, telecontrol):
        self.tc = telecontrol
        self._ensure_python(self.tc.yum)
        self.pip = _Pip(self) 

    def start(self, file):
        return self.tc.start(self._join(file))
    
    
    def _join(self, *s):
        return ' '.join([self.name] + list(s))

    def _is_ok(self):
        names = ['python3', 'python3.4', 'python3.5', 'python3.6']
        for name in names:
            self.tc.sendline(' '.join(['type', name]), stdout=False)
            self.tc.prompt()
            if not self.tc.show():
                self.name = name
                return True
        else:
            return False

    def _ensure_python(self, yum):
        if self._is_ok():
            return
        yum.install('epel-release')
        _, msg = yum.install('python34')
        if self._is_ok():
            return
        raise Exception('\n'.join([
                'Cannot found python.',
                'Messgae from yum:',
                msg]))

class _Pip:

    name = ''

    def __init__(self, python):
        self.python = python
        self.tc = python.tc 

    def __bool__(self):
        return bool(self.name)

    def list(self):
        self._ensure_pip()
        self.tc.sendline(' '.join([self.name, 'list',
                '--format=legacy']), stderr=False)
        self.tc.prompt()
        foo = self.tc.show()
        foo = foo.split('\n')
        foo = filter(lambda x:x, foo)
        foo = [i.split(' ')[0].strip() for i in foo]
        self.tc.reset()
        return foo    
        
    def install(self, packs):
        if not isinstance(packs, list):
            packs = [packs]
        self._ensure_pip()
        for p in packs:
            self.tc.sendline(' '.join([self.name, 'install', p, '-U']), 
                    stdout=False, stderr=False)
            self.tc.prompt()
        foo = self.list()
        for p in packs:
            if p not in foo:
                self.tc.reset()
                return False, p
        else:
            return True, None

        

    def _is_ok(self):
        name = ' '.join([self.python.name, '-m', 'pip'])
        self.tc.sendline(name, stdout=False)
        self.tc.prompt()
        if not self.tc.show():
            self.name = name
            return True
        name = 'pip3'
        self.tc.sendline(name, stdout=False)
        return False

    def _ensure_pip(self):
        if self or self._is_ok():
            return
        self.tc.sendline(' '.join([
                self.python.name, '-m', 'ensurepip']), stdout=False)
        self.tc.prompt()
        msg = self.tc.show()
        if not msg:
            self.tc.sendline(' '.join([
                self.python.name, '-m', 'pip',
                'install', '--upgrade', 'pip']), stdout=False)
            self.tc.prompt()
            msg = self.tc.show()
        if not self._is_ok():
            raise Exception('Pip Error:\n' + msg)

    




import re
import sys
import os
import time

from pexpect.pxssh import pxssh
import pexpect as pe
from pexpect import TIMEOUT
from devtools.format import tostr

from .yum import Yum
from .python import Python

INTERNAL = 0.5

# A wrap for the pxssh Class (which is a wrap of spawn Class).
# Make controlling remote more easily
class Telecontrol(pxssh):
    
    target_addr = ('', 0)

    def __init__(
            self, 
            user, 
            ip='122.144.139.60', 
            port='12045', 
            passwd='yylc#8888', 
            timeout=3,
            echo=True):

        super(Telecontrol, self).__init__(timeout=timeout)
        
        self.login(ip, user, passwd, port=port)
        
        if echo or __debug__ : 
            print('logged in. %s@%s:%s' % (user, ip, port))
        self.target_addr = (ip, str(port))
        self.user = user
        self.passwd = passwd
        
        self.yum = Yum(self)
        self.python = Python(self)
        if __debug__ : print('checked in advance.') 
        if __debug__ : print('Telecontrol created.') 

    def start(self, cmd):
        '''run a cmd with no hup or stdin/out/err. Returns the pid of child process'''
        self.sendline(cmd, nohup=True, stderr=False)
        self.prompt()
        res = re.search('\[\d*\] (\d*)', self.show()).groups()[0]
        return int(res)

    # Override, add sleep to avoid some weird bugs
    def prompt(self, timeout=-1, sleep=0.1):
         time.sleep(sleep)
         res = super(Telecontrol, self).prompt(timeout)
         time.sleep(sleep)
         return res

    # Override, add option stdout=False
    def sendline(self, s='', stdout=True, stderr=True, nohup=False):
        if stdout is False:
            s = s + ' 1>/dev/null'
        if stderr is False:
            s = s + ' 2>/dev/null'
        if nohup is True:
            s = 'nohup ' + s + ' &'
        if nohup:
            if stderr:
                # unfixed !!!!
                raise NotImplementedError('Weird line feed.')
                self.prompt()
                self.expect(os.linesep, timeout=0.1)
                # self.expect(dd[os.linesep], timeout=0.1)
                res = res + self.sendline('clear')
            else:
                res = super(Telecontrol, self).sendline(s) 
                return res

        else: 
            res = super(Telecontrol, self).sendline(s) 
            return res

    # return True:  is a directory
    # return False: is not a directory
    # return None:  no such file or directory
    def isdir(self, filepath):
        self.sendline(filepath + '/')
        self.prompt()
        text = self.readlines()
        if re.search('(?i)is a directory', text):
            return True
        elif re.search('(?i)not a directory', text):
            return False
        elif re.search('(?i)no such file or directory', text):
            return None
        else:
            raise Exception('Unpredicted output:\n {0} \n\n'.format(text))

    # remove file or directory sliently by force
    def remove(self, filepath):
        foo = self.isdir(filepath)
        if foo is True:
            self.sendline('rm -rf %s' % filepath)
            self.prompt()
        elif foo is False:
            self.sendline('rm -f %s' % filepath)
            self.prompt()
        elif foo is None:
            if __debug__: print('Remove {0}: file or dir not found'.format(filepath))
            return 
        if __debug__ : print('Remove {0}'.format(filepath), self.readlines(), sep='\n')

    def mkdir(self, dirpath):
        if not dirpath.startswith('/'):
            raise NotImplementedError('mkdir for relative path not supported yet')
        self.sendline('mkdir %s' % dirpath, stdout=False)
        self.prompt()
        foo = self.readlines()
        if foo:
            return False, foo
        else:
            return True, ''
        
       
    def show(self):
        foo = self.before.decode()
        foo = re.sub('^.*\r\n', '', foo)
        return foo     
    
    def readlines(self):
        foo = self.show()
        foo = foo.split('\n')
        foo = filter(lambda x: x.strip() and not re.match('\[\d*\]\+', x), foo)
        return list(foo)

    def senddir(self, source_dir, target_dir, timeout=10):
        if not os.path.exists(source_dir):
            return False, 'file or directory not found'
        if not os.path.isdir(source_dir):
            raise Exception('Not a dir.')

        target_dir = target_dir + '/'
        self.mkdir(target_dir)
        
        target_path = '{0}@{1}:{2}'.format(
                self.user, 
                self.target_addr[0], 
                target_dir)

        cmd = 'scp -P {0} -r {1}/* {2}'.format(
                self.target_addr[1],
                source_dir,
                target_path)


        child = pe.spawn(cmd, timeout=3)
        time = 0
        
        possibles = [
                '(?i)password: ',  # 0
                '(?i)yes/no',      # 1
                'lost connection', # 2
                pe.EOF,            # 3
                pe.TIMEOUT,        # 4
                'ssh:',            # 5
                '100%',            # 6
                '(?i)no such',     # 7
                ]

        while True:
            index = child.expect(possibles)
            if __debug__ and index != 6: print('Uploading ... code:', index, str(possibles[index]))
            if index == 0:
                child.sendline(self.passwd)
            elif index == 1:
                child.sendline('yes')
            elif index == 2:
                return False, 'lost connection'
            elif index == 3:
                return True, child.before
            elif index == 4:
                time += child.timeout
                if time > timeout:
                    return False, 'upload timeout'
            elif index == 5:
                return False, 'ssh failed'
            elif index == 6:
                pass 
            elif index == 7:
                print('Upload failed, cmd:', cmd)

        

    # This function need to be improved
    def upload(self, source_file, target_dir, target_name='', opt='', timeout=10):
        if not os.path.exists(source_file):
            return False, 'file or directory not found'
        if os.path.isdir(source_file):
            opt = '-r'
        elif os.path.isfile(source_file):
            opt = ''
        else:
            raise Exception('Not a dir nor a file')
        if source_file.strip().endswith('/'):
            raise NotImplementedError('"scp source/dir/ target/dir" not implemented.')
        target_dir = target_dir + '/'
        if not target_name:
            target_name = source_file.split('/')[-1]
        target_path = os.path.join(target_dir, target_name)
        self.mkdir(target_dir)
        if target_dir.count('/') <= 1:
            raise Exception('This is a danger path')
        self.remove(target_path)
        
        target_path = '{0}@{1}:{2}'.format(
                self.user, 
                self.target_addr[0], 
                target_path)
        
         

        cmd = 'scp -P {0} {1} {2} {3}'.format(
                self.target_addr[1],
                opt,
                source_file,
                target_path)

        if __debug__: print('Upload cmd:', cmd)

        child = pe.spawn(cmd, timeout=3)
        time = 0
        
        possibles = [
                '(?i)password: ',  # 0
                '(?i)yes/no',      # 1
                'lost connection', # 2
                pe.EOF,            # 3
                pe.TIMEOUT,        # 4
                'ssh:',            # 5
                '100%']            # 6 
        while True:
            index = child.expect(possibles)
            if __debug__ and index != 6: print('Uploading ... code:', index, str(possibles[index]))
            if index == 0:
                child.sendline(self.passwd)
            elif index == 1:
                child.sendline('yes')
            elif index == 2:
                return False, 'lost connection'
            elif index == 3:
                return True, child.before
            elif index == 4:
                time += child.timeout
                if time > timeout:
                    return False, 'upload timeout'
            elif index == 5:
                return False, 'ssh failed'
            elif index == 6:
                pass 
    def reset(self):
        self.sendcontrol('c')
        self.prompt(0.1)
        self.sendline('clear')
        self.prompt(0.1)
        time.sleep(0.5)
            

            

        



# util for Telecontrol to interact with yum.

from pexpect import TIMEOUT
import os.path as op
import re
import time


dirpath = op.dirname(op.abspath(__file__))

class Yum:
    def __init__(self, telecontrol):
        self.tc = telecontrol
        boo, msg = self._repair()
        if boo is False:
            raise Exception('Yum is not available.')
        
        
    # Internal, blocking
    def _is_ok(self):
        '''check yum is available or not'''
        # @WARNING: This function will never timeout
        self.tc.sendline('yum list yum', stdout=False)
        possibles = [self.tc.PROMPT, TIMEOUT]
        while True:
            index = self.tc.expect(possibles, timeout=5)
            if index == 0:
                if not self.tc.readlines():
                    return True, ''
                return False, self.tc.readlines()
            elif index == 1:
                pass

    # Internal
    def _repair(self):
        '''Try to fix problems on yum'''
        boo, msg = self._is_ok()
        res = 'Unkonw exception when repairing yum.'

        if boo is True:
            return True, ''

        if 'Error: Cannot retrieve metalink for repository: epel.' in msg:
            source = op.join(dirpath, 'epel.repo')
            filepath = '/etc/yum.repos.d/epel2.repo' 
            foo, bar = self.upload(source, filepath)
            if foo is False:
                res = 'Upload repo failed'
            elif __debug__: print('Upload repo successfully.')
        else:
            raise Exception('Unknow yum problem\n' + msg + '\n\n')
        
        boo, msg = self._is_ok()
        if boo is not True:
            return False, res
        else:
            return True, '' 

    def has(self, name):
        '''Find a application installed or not'''
        self.tc.sendline('yum list %s' % name, stdout=False)
        self.tc.prompt(5)
        if not self.tc.readlines():
            return True
        
        name = 'type ' + name

        self.tc.sendline(name, stdout=False)
        self.tc.prompt(5)
    
        foo = self.tc.readlines()
        if foo:
            return False
        return True

    def install(self, name):
        '''Install packages by yum.'''
        if self.has(name) is True:
            return True, 'exist'

        self.tc.sendline('yum install %s -y' % name, stdout=False)
        while True:
            possibles = [TIMEOUT, self.tc.PROMPT]
            index = self.tc.expect(possibles)

            if index == 0:
                pass
            else:
                if not self.tc.readlines():
                    return True, 'complete'
                else:
                    info = '\n'.join(['fail when installing', '='*10, self.tc.readlines(), '='*10, '']) 
                    return False, info
            
    




#!/usr/bin/env python
# coding: utf-8

from setuptools import setup

setup(
    name='AIJIcossh',
    version='0.0.3',
    author='AIJI',
    author_email='thecrazyaiji@gmail.com',
    description='An ssh telecontrol Class for CentOS based on pxssh',
    packages=['cossh'],
    install_requires=['pexpect', 'AIJIdevtools'],
    url='https://github.com/AIJIJI/cossh'
                                                                                )
